__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

from umiche.deduplicate.pipeline.Heterogeneity import Heterogeneity as pipeheter
from umiche.deduplicate.pipeline.Standard import Standard as pipestd
from umiche.recovery.pipeline.Anchor import Anchor


def standard(
        param_fpn : str,
        scenario : str,
        method : str,
        deduped_method : str,
        split_method : str,
        is_collapse_block: bool,
        is_trim : bool = False,
        is_tobam : bool = False,
        is_dedup : bool = False,
        is_sv : bool = False,
        verbose : bool = True,
):
    return pipestd(
        scenario=scenario,
        method=method,
        is_trim=is_trim,
        is_tobam=is_tobam,
        is_dedup=is_dedup,
        is_collapse_block=is_collapse_block,
        deduped_method=deduped_method,
        split_method=split_method,
        param_fpn=param_fpn,
        is_sv=is_sv,
        verbose=verbose,
    )


def heterogeneity(
        param_fpn: str,
        scenario : str,
        method : str,
        is_trim : bool = False,
        is_tobam : bool = False,
        is_dedup : bool = False,
        is_sv : bool = False,
        verbose : bool = False,
):
    return pipeheter(
        param_fpn=param_fpn,
        scenario=scenario,
        method=method,
        is_trim=is_trim,
        is_tobam=is_tobam,
        is_dedup=is_dedup,
        is_sv=is_sv,
        verbose=verbose,
    )


def anchor(
        scenario: str,
        param_fpn: str,
):
    return Anchor(
        scenario=scenario,
        param_fpn=param_fpn,
    ).pct_read_captured()


if __name__ == "__main__":
    from umiche.path import to

    # print(heterogeneity(
    #     # scenario='pcr_nums',
    #     # scenario='pcr_errs',
    #     scenario='seq_errs',
    #     # scenario='ampl_rates',
    #     # scenario='umi_lens',
    #     # scenario='seq_deps',
    #     # scenario='umi_nums',
    #
    #     # method='unique',
    #     # method='cluster',
    #     # method='adjacency',
    #     method='directional',
    #     # method='mcl',
    #     # method='mcl_val',
    #     # method='mcl_ed',
    #     # method='mcl_cc_all_node_umis',
    #     # method='dbscan_seq_onehot',
    #     # method='birch_seq_onehot',
    #     # method='aprop_seq_onehot',
    #     # method='hdbscan_seq_onehot',
    #     # method='set_cover',
    #
    #     # is_trim=True,
    #     # is_tobam=False,
    #     # is_dedup=False,
    #
    #     # is_trim=False,
    #     # is_tobam=True,
    #     # is_dedup=False,
    #
    #     is_trim=False,
    #     is_tobam=False,
    #     is_dedup=True,
    #     is_sv=True,
    #
    #     param_fpn=to('data/params.yml'),
    # ))

    print(standard(
        # scenario='pcr_nums',
        # scenario='pcr_errs',
        scenario='seq_errs',
        # scenario='ampl_rates',
        # scenario='umi_lens',

        # method='unique',
        # method='cluster',
        # method='adjacency',
        # method='directional',
        # method='mcl',
        # method='mcl_val',
        # method='mcl_ed',
        method='set_cover',
        # method='majority_vote',

        # is_trim=True,
        # is_tobam=False,
        # is_dedup=False,

        # is_trim=False,
        # is_tobam=True,
        # is_dedup=False,

        is_trim=False,
        is_tobam=False,
        is_dedup=True,

        # @@ for directional on multimer umis deduplicated by set_cover
        is_collapse_block=False,
        deduped_method='set_cover',
        split_method='split_to_all', # split_to_all split_by_mv

        # @@ for directional on multimer umis deduplicated by majority_vote
        # is_collapse_block=False,
        # deduped_method='majority_vote',
        # split_method='',

        # @@ for directional but on monomer umis of set_cover or majority_vote
        # is_collapse_block=True, # True False
        # collapse_block_method='take_by_order', # majority_vote take_by_order
        # deduped_method='set_cover', # majority_vote set_cover
        # split_method='split_by_mv', # split_to_all split_by_mv

        # @@ for directional but on monomer umis without other methods.
        # is_collapse_block=True,  # True False
        # collapse_block_method='majority_vote',  # majority_vote take_by_order
        # deduped_method='',  # majority_vote set_cover
        # split_method='',  # split_to_all split_by_mv

        # param_fpn=to('data/params_dimer.yml'),
        param_fpn=to('data/params_trimer.yml'),
        # param_fpn=to('data/params.yml'),

        verbose=False, # True False
    ))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from collections import deque
from umiche.util.Console import Console


class ConnectedComponent:

    def __init__(
            self,
            verbose=False,
    ):
        self.console = Console()
        self.console.verbose = verbose

    def deque(
            self,
            graph,
    ):
        """

        Parameters
        ----------
        graph

        Returns
        -------

        """
        visited = set()
        for root, nbrs in graph.items():
            if root not in visited:
                visited.add(root)
                component = []
                queue = deque([root])
                # print('-> root {} has not been visited'.format(root))
                # print('===> a queue built by root {} is {}'.format(root, queue))
                while queue:
                    # print('======> a queue built by each root node {}'.format(queue))
                    node = queue.popleft()
                    # print('======> node: {}'.format(node))
                    component.append(node)
                    for nbr in graph[node]:
                        if nbr not in visited:
                            visited.add(nbr)
                            queue.append(nbr)
                # print('======> visited nodes {}'.format(visited))
                yield component
            else:
                continue
                # print('-> root {} has been visited'.format(root))

    def set(
            self,
            graph,
    ):
        """
        Examples
        --------

        ..  @ex:
            seen = set()
            def component(node):
                nodes = set([node])
                while nodes:
                    node = nodes.pop()
                    seen.add(node)
                    nodes |= neighbors[node] - seen
                    yield node
            for node in neighbors:
                if node not in seen:
                    yield component(node)

        Parameters
        ----------
        graph

        Returns
        -------

        """
        visited = set()
        components = []
        for root, nbrs in graph.items():
            if root not in visited:
                visited.add(root)
                component = []
                queue = [root]
                self.console.print('=========>root {} has not been visited'.format(root))
                self.console.print('=========>a queue built by root {} is {}'.format(root, queue))
                while queue:
                    self.console.print('============>a queue built by each root node {}'.format(queue))
                    node = queue.pop(0)
                    self.console.print('============>node: {}'.format(node))
                    component.append(node)
                    for nbr in graph[node]:
                        if nbr not in visited:
                            visited.add(nbr)
                            queue.append(nbr)
                self.console.print('=========>visited nodes {}'.format(visited))
                components.append(component)
            else:
                self.console.print('=========>root {} has been visited'.format(root))
        return components


if __name__ == "__main__":
    graph_adj = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D', 'G'],
        'F': ['D', 'G'],
        'G': ['E', 'F'],
    }
    p = ConnectedComponent()

    ccs = list(p.deque(graph_adj))

    print(ccs)

    # print(p.set(graph_adj))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from umiche.deepconvcvae.Extreme import extreme as statextreme


class summary(object):

    def __init__(self, ):
        pass

    def extreme(self, cv_dict, gc_mat, type='train', n_top=10):
        std_dict = {}
        for cv_iter, v in cv_dict.items():
            print('===>cv_{}:'.format(cv_iter))
            std_dict[cv_iter] = {}
            extrem_gene_ids = statextreme().perCellClsGeneIds(
                gc_mat=gc_mat,
                cls_vec=v['train_cls'] if type == 'train' else v['test_cls'],
                oriented='maximal',
                n_top=n_top,
                scheme={
                    'method': 'mean',
                },
            )
            retrv_ids = v['train_index'] if type == 'train' else v['test_index']
            gc_mat_cv = gc_mat[retrv_ids, :]
            std_dict[cv_iter]['val_dict_arr_1d'] = {cluster_name: gc_mat_cv[:, sel_gene_ids].flatten() for cluster_name, sel_gene_ids in extrem_gene_ids.items()}
        return std_dict


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
from umiche.util.Console import Console


class adjacency:

    def __init__(
            self,
            graph=None,
            verbose=False,
    ):
        self._graph = graph
        self._graph_mapped = None

        self.console = Console()
        self.console.verbose = verbose

    @property
    def graph(self, ):
        # print('the current graph is {}'.format(self._graph))
        return self._graph

    @graph.setter
    def graph(self, value):
        self._graph = value

    @property
    def graph_mapped(self, ):
        if self._graph == None:
            raise 'go set your graph'
        else:
            return self.map(self._graph)

    @property
    def glen(self, ):
        # print('the number of egdes in the current graph is {}'.format(len(self._graph)))
        return len(self._graph)

    @property
    def key_mapped(self, ):
        return {[*self._graph.keys()][k]: k for k in range(len([*self._graph.keys()]))}

    def map(self, graph):
        self.console.print('============>The graph is being mapped')
        self.console.print('============>key map:\n {}'.format(self.key_mapped))
        if isinstance(graph, dict):
            self.console.print('============>The graph is a dict')
            g_mapped = {}
            for k, vals in self._graph.items():
                g_mapped[self.key_mapped[k]] = []
                for val in vals:
                    g_mapped[self.key_mapped[k]].append(self.key_mapped[val])
            self.console.print('============>The mapped graph is\n {}'.format(g_mapped))
            return g_mapped

    def dict(self, ):
        return self._graph

    def set(self, ):
        adj_set = {}
        for k, vals in self._graph.items():
            adj_set[k] = set(vals)
        return adj_set

    def list(self, ):
        return [*self._graph.values()]

    def matrix(self, ):
        adj_mat = np.zeros(shape=[self.glen, self.glen])
        for k, vals in self._graph.items():
            for val in vals:
                adj_mat[self.key_mapped[k], self.key_mapped[val]] = 1
        return adj_mat

    def hash(self, ):
        return

    def toEdgeList(self, rr=True):
        edges = []
        for k, vals in self._graph.items():
            for val in vals:
                edges.append((k, val))
        if rr:
            repeat = []
            edge_set = set(edges)
            while edges:
                edge = edges.pop(0)
                if tuple(reversed(edge)) in edges:
                    repeat.append(edge)
            edges = list(edge_set.difference(set(repeat)))
        return edges


if __name__ == "__main__":
    graph_adj_dict = {
        'A': ['B', 'C'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['E', 'F'],
        'E': ['D'],
        'F': ['D'],
    }

    # graph_adj_dict = {
    #     0: [4],
    #     1: [2, 5],
    #     2: [1, 3, 8],
    #     3: [2, 14],
    #     4: [0, 9],
    #     5: [1],
    #     6: [10, 12],
    #     7: [13],
    #     8: [2, 14],
    #     9: [4, 10, 15],
    #     10: [6, 9],
    #     11: [17],
    #     12: [6],
    #     13: [7, 19, 20],
    #     14: [3, 8],
    #     15: [9, 21],
    #     16: [22],
    #     17: [11, 18],
    #     18: [17, 19],
    #     19: [13, 18, 26],
    #     20: [13, 26],
    #     21: [15, 27],
    #     22: [16, 23],
    #     23: [22, 24, 28],
    #     24: [23, 25, 29],
    #     25: [24],
    #     26: [19, 20, 30, 31],
    #     27: [21],
    #     28: [23, 29],
    #     29: [24, 28],
    #     30: [26, 31],
    #     31: [26, 30],
    # }

    p = adjacency(graph_adj_dict)

    p.graph = p.graph_mapped

    print(p.graph)

    print(p.dict())

    print(p.set())

    print(p.list())

    print(p.matrix())

    print(p.toEdgeList(rr=False))
    print(p.toEdgeList(rr=True))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import pandas as pd
from sklearn.cluster import HDBSCAN as skhdbscan
from sklearn.cluster import DBSCAN as skdbscan
from sklearn.cluster import Birch as skbirch
from sklearn.cluster import AffinityPropagation as skaprop

from umiche.deduplicate.method.ReformKit import ReformKit as refkit
from umiche.network.Adjacency import Adjacency as netadj
from umiche.util.Console import Console


class Clustering:

    def __init__(
            self,
            clustering_method='dbscan',
            heterogeneity=None,
            verbose=True,
            **kwargs
    ):
        self.refkit = refkit()
        self.netadj = netadj()

        self.clustering_method = clustering_method
        self.heterogeneity = heterogeneity

        self.kwargs = kwargs
        if 'dbscan_eps' in self.kwargs.keys():
            self.dbscan_eps = self.kwargs['dbscan_eps']
        else:
            self.dbscan_eps = None
        if 'dbscan_min_spl' in self.kwargs.keys():
            self.dbscan_min_spl = self.kwargs['dbscan_min_spl']
        else:
            self.dbscan_min_spl = None
        if 'birch_thres' in self.kwargs.keys():
            self.birch_thres = self.kwargs['birch_thres']
        else:
            self.birch_thres = None
        if 'birch_n_clusters' in self.kwargs.keys():
            self.birch_n_clusters = self.kwargs['birch_n_clusters']
        else:
            self.birch_n_clusters = None
        if 'hdbscan_min_spl' in self.kwargs.keys():
            self.hdbscan_min_spl = self.kwargs['hdbscan_min_spl']
        else:
            self.hdbscan_min_spl = None
        if 'aprop_preference' in self.kwargs.keys():
            self.aprop_preference = self.kwargs['aprop_preference']
        else:
            self.aprop_preference = None
        if 'aprop_random_state' in self.kwargs.keys():
            self.aprop_random_state = self.kwargs['aprop_random_state']
        else:
            self.aprop_random_state = None

        self.console = Console()
        self.console.verbose = verbose

    @property
    def tool(self, ):
        return {
            'dbscan': skdbscan(eps=self.dbscan_eps, min_samples=self.dbscan_min_spl),
            'birch': skbirch(threshold=self.birch_thres, n_clusters=None),
            'hdbscan': skhdbscan(min_samples=self.hdbscan_min_spl),
            'aprop': skaprop(preference=self.aprop_preference, random_state=self.aprop_random_state),
        }

    def adj_to_edge_list(self, graph):
        self.netadj.graph = graph
        return self.netadj.to_edge_list()

    def tovertex(self, x):
        clustering_cluster_arr = x['clustering_clusters'][0]
        cc_vertex_arr = x['cc_vertices']
        uniq_cls_arr = np.unique(clustering_cluster_arr)
        clusters = [[] for _ in range(len(uniq_cls_arr))]
        for i, cls in enumerate(clustering_cluster_arr):
            clusters[cls].append(cc_vertex_arr[i])
        return clusters

    def dfclusters(
            self,
            connected_components,
            graph_adj,
            int_to_umi_dict,
    ):
        """

        Parameters
        ----------
        connected_components
            connected components in dict format:
            {
                'cc0': [...] # nodes,
                'cc1': [...],
                'cc2': [...],
                ...
                'ccn': [...],
            }
            e.g.
            {
                0: ['A', 'B', 'C', 'D', 'E', 'F'],
            }
        graph_adj
            the adjacency list of a graph

        Returns
        -------
            a pandas dataframe
            each connected component is decomposed into more connected subcomponents.

        """
        ### @@ graph_cc_adj
        # {'A': ['B', 'C', 'D'], 'B': ['A', 'C'], 'C': ['A', 'B'], 'D': ['A'], 'E': ['G'], 'F': ['G'], 'G': ['E', 'F']}

        # When an adjacency list of a graph is shown as above, we have the output in the following.
        df_ccs = pd.DataFrame({'cc_vertices': [*connected_components.values()]})
        ### @@ df_ccs
        #     cc_vertices
        # 0  [A, B, C, D]
        # 1     [E, G, F]
        df_ccs['umi'] = df_ccs['cc_vertices'].apply(lambda x: [int_to_umi_dict[node] for node in x])
        ### @@ df_ccs['umi']
        # 0    [AGATCTCGCA, AGATCCCGCA, AGATCACGCA, AGATCGCGCA]
        # 1                [AGATCGCGGA, TGATCGCGAA, AGATCGCGTA]
        # Name: umi, dtype: object
        df_ccs['onehot'] = df_ccs['umi'].apply(lambda umi_arr: [self.refkit.onehot(umi=umi) for umi in umi_arr])
        ### @@ df_ccs['onehot']
        # 0    [[1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0,...
        # 1    [[1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0,...
        # Name: onehot, dtype: object

        clustering_ins = self.tool[self.clustering_method]
        df_ccs['clustering_clusters'] = df_ccs['onehot'].apply(lambda onehot_2d_arrs: [
            clustering_ins.fit(onehot_2d_arrs).labels_
        ])
        # print(df_ccs['clustering_clusters'])
        df_ccs['clusters'] = df_ccs.apply(lambda x: self.tovertex(x), axis=1)
        # print(df_ccs['clusters'])
        df_ccs['clust_num'] = df_ccs['clusters'].apply(lambda x: len(x))
        # print(df_ccs['clust_num'])
        df_ccs['graph_cc_adj'] = df_ccs['cc_vertices'].apply(lambda x: self.refkit.graph_cc_adj(x, graph_adj))
        # print(df_ccs['graph_cc_adj'])
        df_ccs['edge_list'] = df_ccs['graph_cc_adj'].apply(lambda graph: self.adj_to_edge_list(graph=graph))
        # print(df_ccs['edge_list'])
        df_ccs['apv'] = df_ccs['edge_list'].apply(lambda edge_list: [list(el) for el in edge_list])
        # print(df_ccs['apv'])
        return df_ccs

    def dfclusters_adj_mat(
            self,
            connected_components,
            graph_adj,
    ):
        """

        Parameters
        ----------
        connected_components
            connected components in dict format:
            {
                'cc0': [...] # nodes,
                'cc1': [...],
                'cc2': [...],
                ...
                'ccn': [...],
            }
            e.g.
            {
                0: ['A', 'B', 'C', 'D', 'E', 'F'],
            }
        graph_adj
            the adjacency list of a graph

        Returns
        -------
            a pandas dataframe
            each connected component is decomposed into more connected subcomponents.

        """
        ### @@ graph_cc_adj
        # {'A': ['B', 'C', 'D'], 'B': ['A', 'C'], 'C': ['A', 'B'], 'D': ['A'], 'E': ['G'], 'F': ['G'], 'G': ['E', 'F']}

        # When an adjacency list of a graph is shown as above, we have the output in the following.
        df_ccs = pd.DataFrame({'cc_vertices': [*connected_components.values()]})
        ### @@ df_ccs
        #     cc_vertices
        # 0  [A, B, C, D]
        # 1     [E, G, F]
        df_ccs['graph_cc_adj'] = df_ccs['cc_vertices'].apply(lambda x: self.refkit.graph_cc_adj(x, graph_adj))
        # print(df_ccs['graph_cc_adj'])
        df_ccs['cc_adj_mat'] = df_ccs.apply(
            # lambda x: self.matrix(
            #     graph_adj=x['graph_cc_adj'],
            #     key_map=x['nt_to_int_map'],
            # ),
            lambda x: netadj(graph=x['graph_cc_adj']).to_matrix(),
            axis=1,
        )
        # print(df_ccs['cc_adj_mat'])
        clustering_ins = self.tool[self.clustering_method]
        df_ccs['clustering_clusters'] = df_ccs['cc_adj_mat'].apply(lambda adj_mat: [
            clustering_ins.fit(adj_mat).labels_
        ])
        # print(df_ccs['clustering_clusters'])
        df_ccs['clusters'] = df_ccs.apply(lambda x: self.tovertex(x), axis=1)
        # print(df_ccs['clusters'])
        df_ccs['clust_num'] = df_ccs['clusters'].apply(lambda x: len(x))
        # print(df_ccs['clust_num'])
        df_ccs['edge_list'] = df_ccs['graph_cc_adj'].apply(lambda graph: self.adj_to_edge_list(graph=graph))
        # print(df_ccs['edge_list'])
        df_ccs['apv'] = df_ccs['edge_list'].apply(lambda edge_list: [list(el) for el in edge_list])
        # print(df_ccs['apv'])
        return df_ccs

    def dfclusters_cc_fuse(
            self,
            connected_components,
            df_umi_uniq_val_cnt,
            int_to_umi_dict,
    ):
        df_ccs = pd.DataFrame({'cc_vertices': [*connected_components.values()]})
        print(df_ccs)
        ### @@ df_ccs
        #              cc_vertices
        # 0  [A, B, C, D, E, F, G]
        df_ccs['cc_max_id'] = df_ccs['cc_vertices'].apply(lambda cc: self.refkit.maxid(df_umi_uniq_val_cnt, cc))
        ### @@ df_ccs['cc_max_id']
        # 0    A
        # Name: cc_max_id, dtype: object
        df_ccs['cc_max_id2seq'] = df_ccs['cc_max_id'].apply(lambda x: int_to_umi_dict[x])
        print(df_ccs['cc_max_id2seq'])
        # 0    AGATCTCGCA
        # Name: cc_max_id2seq, dtype: object
        vertex_onehot = df_ccs['cc_max_id2seq'].apply(lambda umi: self.refkit.onehot(umi=umi)).values.tolist()
        df_vertex_onehot = pd.DataFrame(vertex_onehot)
        print(df_vertex_onehot)
        # d = skdbscan(eps=2.5, min_samples=1).fit(df_vertex_onehot)
        d =  skbirch(threshold=1.8, n_clusters=None).fit(df_vertex_onehot)
        asd = np.unique(d.labels_)
        asdas = np.array(d.labels_)
        labels = d.labels_
        print(d.labels_)
        print(asd)
        print(asdas)
        n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
        n_noise_ = list(labels).count(-1)
        print(n_clusters_)
        print(n_noise_)
        # return len(asd)
        return len(asd) + len(asdas[asdas == -1]), len(asdas[asdas == -1])

    def dfclusters_cc_all_node_umis(
            self,
            graph_adj,
            int_to_umi_dict,
    ):
        """

        Parameters
        ----------
        connected_components
            connected components in dict format:
            {
                'cc0': [...] # nodes,
                'cc1': [...],
                'cc2': [...],
                ...
                'ccn': [...],
            }
            e.g.
            {
                0: ['A', 'B', 'C', 'D', 'E', 'F'],
            }
        graph_adj
            the adjacency list of a graph

        Returns
        -------
            a pandas dataframe
            each connected component is decomposed into more connected subcomponents.

        """
        df_ccs = pd.DataFrame()
        df_ccs.loc[0, 'method'] = 'dfclusters_cc_all_node_umis'
        onehot_2d_arrs = [list(self.refkit.onehot(umi=umi)) for k, umi in int_to_umi_dict.items()]
        d1 = {'dfclusters_cc_all_node_umis': onehot_2d_arrs}
        d2 = {'dfclusters_cc_all_node_umis': [*int_to_umi_dict.keys()]}
        df_ccs['onehot'] = df_ccs['method'].apply(lambda x: d1[x])
        df_ccs['cc_vertices'] = df_ccs['method'].apply(lambda x: d2[x])
        # print(df_ccs)
        clustering_ins = self.tool[self.clustering_method]
        df_ccs['clustering_clusters'] = df_ccs['onehot'].apply(lambda onehot_2d_arrs: [
            clustering_ins.fit(onehot_2d_arrs).labels_
        ])
        # print(df_ccs['clustering_clusters'])
        df_ccs['clusters'] = df_ccs.apply(lambda x: self.tovertex(x), axis=1)
        # print(df_ccs['clusters'])
        df_ccs['clust_num'] = df_ccs['clusters'].apply(lambda x: len(x))
        # print(df_ccs['clust_num'])
        df_ccs['graph_cc_adj'] = df_ccs['cc_vertices'].apply(lambda x: self.refkit.graph_cc_adj(x, graph_adj))
        # print(df_ccs['graph_cc_adj'])
        df_ccs['edge_list'] = df_ccs['graph_cc_adj'].apply(lambda graph: self.adj_to_edge_list(graph=graph))
        # print(df_ccs['edge_list'])
        df_ccs['apv'] = df_ccs['edge_list'].apply(lambda edge_list: [list(el) for el in edge_list])
        # print(df_ccs['apv'])
        return df_ccs

    def decompose(
            self,
            list_nd,
    ):
        """

        Parameters
        ----------
        df

        Returns
        -------
        {

        }

        """
        # print(list_nd)
        list_md = []
        for i in list_nd:
            list_md = list_md + i
        res = {}
        for i, cc_sub_each_mcl in enumerate(list_md):
            res[i] = cc_sub_each_mcl
        # print(res)
        return res


if __name__ == "__main__":
    from umiche.deduplicate.method.Cluster import Cluster as umimonoclust

    ### @@ data from UMI-tools
    # graph_adj = {
    #     'A': ['B', 'C', 'D'],
    #     'B': ['A', 'C'],
    #     'C': ['A', 'B'],
    #     'D': ['A', 'E', 'F'],
    #     'E': ['D'],
    #     'F': ['D'],
    # }
    # print("An adjacency list of a graph:\n{}".format(graph_adj))
    #
    # node_val_sorted = pd.Series({
    #     'A': 456,
    #     'E': 90,
    #     'D': 72,
    #     'B': 2,
    #     'C': 2,
    #     'F': 1,
    # })
    # print("Counts sorted:\n{}".format(node_val_sorted))

    ### @@ data from mine
    # graph_adj = {
    #     'A': ['B', 'C', 'D'],
    #     'B': ['A', 'C'],
    #     'C': ['A', 'B'],
    #     'D': ['A', 'E', 'F'],
    #     'E': ['D', 'G'],
    #     'F': ['D', 'G'],
    #     'G': ['E', 'F'],
    # }
    graph_adj = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A',],
        'E': ['G'],
        'F': ['G'],
        'G': ['E', 'F'],
    }
    print("An adjacency list of a graph:\n{}".format(graph_adj))

    node_val_sorted = pd.Series({
        'A': 120,
        'D': 90,
        'E': 50,
        'G': 5,
        'B': 2,
        'C': 2,
        'F': 1,
    })
    print("Counts sorted:\n{}".format(node_val_sorted))

    int_to_umi_dict = {
        'A': 'AGATCTCGCA',
        'B': 'AGATCCCGCA',
        'C': 'AGATCACGCA',
        'D': 'AGATCGCGCA',
        'E': 'AGATCGCGGA',
        'F': 'AGATCGCGTA',
        'G': 'TGATCGCGAA',
    }

    ccs = umimonoclust().cc(graph_adj=graph_adj)
    print("Connected components:\n{}".format(ccs))

    p = Clustering(
        clustering_method='dbscan',
        dbscan_eps=1.5,
        dbscan_min_spl=1,
        birch_thres=1.8,
        birch_n_clusters=None,
    )

    df = p.dfclusters(
        connected_components=ccs,
        graph_adj=graph_adj,
        # df_umi_uniq_val_cnt=node_val_sorted,
        int_to_umi_dict=int_to_umi_dict,
    )
    print(df)
    df_decomposed = p.decompose(list_nd=df['clusters'].values)
    print("deduplicated clusters decomposed:\n{}".format(df_decomposed))

    # df = p.dfclusters_adj_mat(
    #     connected_components=ccs,
    #     graph_adj=graph_adj,
    # )
    # print(df)
    # df_decomposed = p.decompose(list_nd=df['clusters'].values)
    # print("deduplicated clusters decomposed:\n{}".format(df_decomposed))

    # df = p.dfclusters_cc_all_node_umis(
    #     graph_adj=graph_adj,
    #     int_to_umi_dict=int_to_umi_dict,
    # )
    # print(df)
    # df_decomposed = p.decompose(list_nd=df['clusters'].values)
    # print("deduplicated clusters decomposed:\n{}".format(df_decomposed))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import pandas as pd
from DimensionReduction import dimensionReduction as drplot


def read(
        gmat_fpn,
):
    df_gmat = pd.read_hdf(gmat_fpn, 'df')
    y = df_gmat['Y'].values
    if 'Y' in df_gmat.columns:
        df_gmat = df_gmat.drop(columns=['Y'])
    return df_gmat, y


title = '50 epoch 2000 cell'
# title = '50 epoch 1000 cell-new'
# title = '100 epoch 5000 cell'

df, y = read(
    gmat_fpn=to('tresor/') + title + '.h5',
)
x = df.values
print(123123123213, x.shape)
# print(np.unique(df.to_numpy().flatten()).tolist())
# print(x)
# print(y)

# condi = 'umi_nums_seq_err_0.1'
# condi = 'umi_nums_seq_err_0.05'
# condi = 'umi_nums_seq_err_0.01'
condi = 'umi_nums_seq_err_0.005'

direc_dedup = pd.read_csv(to('mclumi/') + condi + '/directional_dedup.txt', header=0)
# print(direc_dedup)

# print(np.unique(df.to_numpy().flatten()).tolist())
uniq = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45]

dedup_direc_map = {x:y for x, y in zip(uniq, direc_dedup['pn0'].values.tolist())}
# print(dedup_direc_map)
x_direc = df.replace(uniq, direc_dedup['pn0'].values.tolist()).values


# drplot().single(X=x, y=y, tech='PCA', marker_size=1, cmap='tab20b', title=title)
# drplot().single(X=x_direc, y=y, tech='PCA', marker_size=1, cmap='tab20b', title=title)

# drplot().single(X=x, y=y, tech='UMAP', marker_size=1, cmap='tab20b', title=title)
# drplot().single(X=x_direc, y=y, tech='UMAP', marker_size=1, cmap='tab20b', title=title)

drplot().single(X=x, y=y, tech='TSNE', marker_size=1, cmap='tab20b', title=title) # Set3 tab20b
drplot().single(X=x_direc, y=y, tech='TSNE', marker_size=1, cmap='tab20b', title=title) # Set3 tab20b




__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL v3.0"
__email__ = "jianfeng.sunmt@gmail.com"
__maintainer__ = "Jianfeng Sun"


class Touch:

    # def __init__(self, task):
    #     self.task = task

    def bar_height_label(
            self,
            canvas,
            ax,
            format,
            fontsize=8,
    ):
        """
        Attach a text label above each bar in *rects*, displaying its height.

        Parameters
        ----------
        canvas
        ax
        format
        fontsize

        Returns
        -------

        """
        heights = []
        for rect in canvas:
            height = rect.get_height()
            heights.append(height)
            # print(height)
            if format == 'pct':
                val_formatted = '{:.2%}'.format(height)
            elif format == 'int':
                val_formatted = '{:d}'.format(int(height))
            else:
                val_formatted = '{}'.format(height)
            ax.annotate(
                val_formatted,
                xy=(rect.get_x() + rect.get_width() / 2, height),
                xytext=(0, 3),  # 3 points vertical offset
                textcoords="offset points",
                ha='center',
                va='bottom',
                fontsize=fontsize,
            )
        return max(heights)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import textwrap
import numpy as np
import pandas as pd
from simreadflow.util.sequence.fastq.Read import read as rfastq
from simreadflow.util.sequence.fastq.Write import write as wfastq
from simreadflow.util.random.Number import number as rannum
from simreadflow.util.file.read.Reader import reader as gfreader
from simreadflow.util.file.write.Writer import writer as fwriter
from umiche.trim.Template import template as umitrim
from umiche.trim.Reader import reader as trimreader
from umiche.deduplicate.dimer.pipeline import Config
from umiche.path import to


class selfHealing(Config.config):

    def __init__(self, fastq_fp, cat):
        super(selfHealing, self).__init__()
        self.fastq_fp = fastq_fp
        self.cat = cat
        self.umitrim = umitrim
        self.gfreader = gfreader()
        self.rfastq = rfastq()
        self.wfastq = wfastq()
        self.trimreader = trimreader()
        self.rannum = rannum()
        self.fwriter = fwriter()

    def rea(self, ):
        for i_pn in range(self.permutation_num):
            for id, i_seq_err in enumerate(self.seq_errs):
                read_stime = time.time()
                names, seqs, _, _ = self.rfastq.fromgz(
                    fastq_path=self.fastq_fp + 'seq_errs/permute_' + str(i_pn) + '/trimmed/',
                    fastq_name=self.cat + '_' + str(id),
                    method='pyfastx',
                )
                print('permutation {}, No.{} with criterion: {}'.format(i_pn, id, i_seq_err))
                # df = self.trimreader.todf(names=names, seqs=seqs)
                df = self.trimreader.todfFromTree(names=names, seqs=seqs)
                # print(df)
                df['origin_info'] = df.apply(lambda x: str(x['umi#']) + '_' + x['umi_src'], axis=1)
                mono_corr_stime = time.time()
                df['umi_monos'] = df['umi'].apply(lambda x: self.split(x, num=20))
                df['umi_correct'] = df['umi'].apply(lambda x: self.correct(x))
                df['umi_mark'] = df['umi_monos'].apply(lambda x: self.marker(x))
                df['umi_l'] = df['umi_monos'].apply(lambda x: x.split(';')[0])
                df['umi_r'] = df['umi_monos'].apply(lambda x: x.split(';')[1])
                df['to_fas'] = df.apply(lambda x: x['origin_info'] + '_' + x['umi_l'], axis=1)
                df['to_fas_correct'] = df.apply(lambda x: x['origin_info'] + '_' + x['umi_correct'], axis=1)
                df_fastq_cp = df.copy()
                df_fastq_cp['to_fas'] = df_fastq_cp.apply(lambda x: x['origin_info'] + '_' + x['umi_r'], axis=1)
                df_umi_extra = df.loc[df['umi_mark'] == 1]
                df_umi_extra['to_fas'] = df_umi_extra.apply(lambda x: x['origin_info'] + '_' + x['umi_r'], axis=1)
                # print(df_umi_extra['to_fas'])
                df_merge = pd.concat(
                    [df[['seq_raw', 'to_fas']], df_fastq_cp[['seq_raw', 'to_fas']]],
                    axis=0,
                ).reset_index(drop=True)
                print(df_merge)
                df_extra = pd.concat(
                    [df[['seq_raw', 'to_fas']], df_umi_extra[['seq_raw', 'to_fas']]],
                    axis=0,
                ).reset_index(drop=True)
                print(df_extra)
                self.wfastq.togz(
                    list_2d=df[['seq_raw', 'to_fas']].values,
                    sv_fp=self.fastq_fp + 'seq_errs/permute_' + str(i_pn) + '/ref/',
                    fn=self.cat + '_' + str(id),
                )
                self.wfastq.togz(
                    list_2d=df[['seq_raw', 'to_fas_correct']].values,
                    sv_fp=self.fastq_fp + 'seq_errs/permute_' + str(i_pn) + '/correct/',
                    fn=self.cat + '_' + str(id),
                )
                self.wfastq.togz(
                    list_2d=df_merge.values,
                    sv_fp=self.fastq_fp + 'seq_errs/permute_' + str(i_pn) + '/double/',
                    fn=self.cat + '_' + str(id),
                )
                self.wfastq.togz(
                    list_2d=df_extra.values,
                    sv_fp=self.fastq_fp + 'seq_errs/permute_' + str(i_pn) + '/bipartite/',
                    fn=self.cat + '_' + str(id),
                )
                print('===>monos time: {:.3f}s'.format(time.time() - mono_corr_stime))
        return

    def split(self, umi, num=24):
        vernier = [i for i in range(num) if i % 2 == 0]
        umi_dimers = [umi[v: v + 2] for v in vernier]
        l = []
        r = []
        for umi_dimer in umi_dimers:
            l.append(umi_dimer[0])
            r.append(umi_dimer[1])
        l = ''.join(l)
        r = ''.join(r)
        # print(l)
        # print(r)
        return l + ';' + r

    def correct(self, umi):
        umi_dimers = textwrap.wrap(umi, 2)
        t = []
        # print(umi_dimers)
        for umi_dimer in umi_dimers:
            if umi_dimer[0] != umi_dimer[1]:
                rand_index = np.random.randint(low=0, high=2, size=1)[0]
                # print(rand_index)
                t.append(umi_dimer[rand_index])
            else:
                t.append(umi_dimer[0])
        return ''.join(t)

    def marker(self, umi):
        tt = umi.split(';')
        if tt[0] != tt[1]:
            return 1
        else:
            return 0


if __name__ == "__main__":
    p = selfHealing(
        fastq_fp=to('data/simu/dimer/treepcr22_250/'),
        cat='seq_err',
    )
    print(p.rea())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import os


class Folder:

    def osmkdir(self, DIRECTORY):
        """

        Parameters
        ----------
        DIRECTORY

        Returns
        -------

        """
        if not os.path.exists(DIRECTORY):
            os.makedirs(DIRECTORY)
        return 0



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from umiche.deepconvcvae.Extreme import extreme as statextreme


class summary:

    def __init__(self, ):
        pass

    def extreme(self, gc_mat, cls_vec, n_top=10):
        print('===>train_sum_extreme')
        extrem_gene_ids = statextreme().perCellClsGeneIds(
            gc_mat=gc_mat,
            cls_vec=cls_vec,
            oriented='maximal',
            n_top=n_top,
            scheme={
                'method': 'mean',
            },
        )
        return {
            cluster_name: gc_mat[:, sel_gene_ids].flatten() for cluster_name, sel_gene_ids in extrem_gene_ids.items()
        }



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import os
import gzip
from umiche.util.Folder import Folder as crtfolder


class Writer:

    def __init__(self):
        pass

    def togz(self, list_2d, sv_fpn, symbol='_'):
        """

        Parameters
        ----------
        list_2d
        sv_fpn
        symbol

        Returns
        -------

        """
        crtfolder().osmkdir(DIRECTORY=os.path.dirname(sv_fpn))
        f = gzip.open(sv_fpn, 'wt')
        for i, read in enumerate(list_2d):
            seq = str(read[0])
            # print(seq)
            # print('@' + symbol.join(read[1:]) + '\n')
            # print('No.{} saving in FASTQ format.'.format(i + 1))
            f.write('@' + symbol.join(read[1:]) + '\n')
            f.write(seq + '\n')
            f.write('+' + '\n')
            f.write('s' * len(seq) + '\n')
        f.close()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
from umiche.util.Console import Console


class Edge:

    def __init__(
            self,
            graph=None,
            verbose=False,
    ):
        self._graph = graph
        self._graph_mapped = None

        self.console = Console()
        self.console.verbose = verbose

    @property
    def graph(self, ):
        self.console.print('======>the current graph is {}'.format(self._graph))
        return self._graph

    @property
    def glen(self, ):
        self.console.print('======>the number of egdes in the current graph is {}'.format(len(self._graph)))
        return len(self._graph)

    @property
    def nodes(self, ):
        g_np = np.array(self._graph)
        # print(self._graph)
        if g_np.size != 0:
            l = np.array(g_np)[:, 0]
            r = np.array(g_np)[:, 1]
            return list(np.unique(np.concatenate((l, r))))
        else:
            return []

    @property
    def key_mapped(self, ):
        return {k: id for id, k in enumerate(self.nodes)}

    @graph.setter
    def graph(
            self,
            value,
    ):
        """

        Parameters
        ----------
        value

        Returns
        -------

        """
        self._graph = value

    @property
    def graph_mapped(self, ):
        if self._graph == None:
            raise 'please set your graph using an edge list'
        else:
            return self.map(self._graph)

    @property
    def rvredanduncy(self, ):
        repeat = []
        edges = self._graph.copy()
        edge_set = set(self._graph)
        while edges:
            edge = edges.pop(0)
            if tuple(reversed(edge)) in edges:
                repeat.append(edge)
        edges = list(edge_set.difference(set(repeat)))
        # print(self._graph)
        return edges

    def map(
            self,
            graph,
    ):
        """

        Parameters
        ----------
        graph

        Returns
        -------

        """
        # print('===>the graph is being mapped')
        # print('======>key map: {}'.format(self.key_mapped))
        g_mapped = []
        for i, edge in enumerate(graph):
            g_mapped.append((self.key_mapped[edge[0]], self.key_mapped[edge[1]]))
        # print('======>the mapped graph: {}'.format(g_mapped))
        return g_mapped

    def to_adj_dict(self, ):
        adj_list = {}
        # scan the arrays edge_u and edge_v
        # print('nodes are {}'.format(self.nodes))
        for i in self.nodes:
            adj_list[i] = []
        for i in range(self.glen):
            l = self._graph[i][0]
            r = self._graph[i][1]
            adj_list[l].append(r)
            adj_list[r].append(l)
        # print(adj_list)
        return adj_list

    def fromlist(
            self,
            list_2d,
    ):
        """

        Parameters
        ----------
        list_2d

        Returns
        -------

        """
        tuples = []
        for e in list_2d:
            tuples.append((e[0], e[1]))
        return tuples

    def fromdf(
            self,
            df,
            col_name1='node_1',
            col_name2='node_2',
            to_tuple=False,
    ):
        """

        Parameters
        ----------
        df
        col_name1
        col_name2
        to_tuple

        Returns
        -------

        """
        edge_list = df[[col_name1, col_name2]].values.tolist()
        if to_tuple:
            return self.fromlist(edge_list)
        else:
            return edge_list


if __name__ == "__main__":
    graph_edge_list = [
        ('A', 'B'),
        ('A', 'C'),
        ('B', 'C'),
        ('A', 'D'),
        ('D', 'E'),
        ('D', 'F'),
        ('D', 'F'),
    ]

    p = Edge(graph_edge_list)

    # p.graph = p.graph_mapped

    print(p.graph)
    print(p.key_mapped)

    p.graph = p.rvredanduncy

    print(p.graph)
    print(p.key_mapped)

    # print(p.rvredanduncy)

    print(p.to_adj_dict())
    print(p.graph_mapped)

    # print(p.nodes)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import sys
import pandas as pd

from umiche.bam.Reader import Reader as alireader

from umiche.bam.Build import Build as umibuild
from umiche.deduplicate.Gadgetry import Gadgetry as umigadgetry
from umiche.deduplicate.Tabulate import Tabulate as umitab

# dedup methods
from umiche.deduplicate.method.Cluster import Cluster as umiclust

from umiche.util.Console import Console
sys.setrecursionlimit(15000000)


class Gene:

    def __init__(
            self,
            bam_fpn,
            ed_thres,
            gene_assigned_tag,
            gene_is_assigned_tag,
            umi_='_',
            work_dir='./',

            heterogeneity=False,
            verbose=False,

            **kwargs,
    ):
        """

        Parameters
        ----------
        bam_fpn
            str - the full path of a BAM file curated by requirements of different dedup modules
        ed_thres
            int - an edit distance threshold (>1, integer)
        method
            str - a deduplication method (mcl, mcl_val, mcl_ed, cluster, unique, ajacency, directional)
        mode
            str - externally or internally run the module (external by defualt, internal)
        mcl_fold_thres
            float - a mcl fold threshold (1.5 by defualt)
        inflat_val
            float - an inflation value for generating mcl clusters (2.0 by defualt)
        exp_val
            int - an expansion value for generating mcl clusters (2 by defualt)
        iter_num
            int - number of iterations for mcl (100 by defualt)
        sv_fpn
            str - the deduplication file path
        verbose
            bool - print log on the console, (True by default or False)
        """
        self.bam_fpn = bam_fpn
        self.ed_thres = ed_thres
        self.gene_assigned_tag = gene_assigned_tag
        self.gene_is_assigned_tag = gene_is_assigned_tag
        self.work_dir = work_dir
        self.heterogeneity = heterogeneity
        self.verbose = verbose
        self.kwargs = kwargs

        self.umiclust = umiclust()

        self.umibuild = umibuild
        self.umigadgetry = umigadgetry()
        self.console = Console()
        self.console.verbose = self.verbose

        # sys.stdout = open(self.work_dir + self.method + '_log.txt', 'w')

        self.alireader = alireader(bam_fpn=self.bam_fpn, verbose=self.verbose)
        self.df_bam = self.alireader.todf(tags=[self.gene_assigned_tag, self.gene_is_assigned_tag])
        ### @@ self.df_bam
        #           id  ... query_qualities
        # 0          0  ...            None
        # 1          1  ...            None
        # 2          2  ...            None
        # ...      ...  ...             ...
        # 20683  20683  ...            None
        self.console.print('======># of raw reads: {}'.format(self.df_bam.shape[0]))
        self.df_bam = self.df_bam.loc[self.df_bam['reference_id'] != -1]
        self.console.print('======># of reads with qualified chrs: {}'.format(self.df_bam.shape[0]))
        self.df_bam = self.df_bam.loc[self.df_bam[self.gene_is_assigned_tag] == 'Assigned']

        # self.df_bam['umi'] = self.df_bam['query_name'].apply(lambda x: x.split(umi_)[1])
        self.df_bam['umi'] = self.df_bam['query_name'].apply(lambda x: x.split(umi_)[-1])
        self.console.print('======># of unique umis: {}'.format(self.df_bam['umi'].unique().shape[0]))
        self.console.print('======># of redundant umis: {}'.format(self.df_bam['umi'].shape[0]))
        self.console.print('======>edit distance thres: {}'.format(self.ed_thres))

        ### @@ self.df_bam
        #           id                     query_name  ...        umi  source
        # 0          0   SRR2057595.2985267_ACCGGTTTA  ...  ACCGGTTTA       1
        # 1          1  SRR2057595.13520751_CCAGGTTCT  ...  CCAGGTTCT       1
        # 2          2   SRR2057595.8901432_AGCGGTTAC  ...  AGCGGTTAC       1
        # ...      ...                            ...  ...        ...     ...
        # 20683  20683  SRR2057595.11966225_ACCGGTTGG  ...  ACCGGTTGG       1
        # [20684 rows x 13 columns]
        self.df_bam_gp = self.df_bam.groupby(by=[self.gene_assigned_tag])
        self.pos_gp_keys = self.df_bam_gp.groups.keys()

        self.console.print('======># of columns in the bam df: {}'.format(len(self.df_bam.columns)))
        self.console.print('======>Columns in the bam df: {}'.format(self.df_bam.columns.tolist()))
        self.console.print('======># of raw reads:\n{}'.format(self.df_bam))

        self.console.print('===>start building umi graphs...')
        umi_graph_build_stime = time.time()
        pos_gps = []
        res_sum = []
        for pos_g in self.pos_gp_keys:
            umi_vignette = self.umibuild(
                df=self.df_bam_gp.get_group(pos_g),
                ed_thres=self.ed_thres,
                verbose=verbose,
            ).data_summary
            # print(umi_vignette)
            # if len(umi_vignette['int_to_umi_dict']) == 1:
            #     print(True)
            #     print(umi_vignette['int_to_umi_dict'])
            #     continue
            # else:
            cc = self.umiclust.cc(umi_vignette['graph_adj'])
            # import json
            # with open('data.json', 'w') as f:
            #     json.dump(cc, f)
            pos_gps.append(pos_g)
            res_sum.append([
                umi_vignette,
                cc,
                umi_vignette['ave_ed'],
                [*umi_vignette['int_to_umi_dict'].keys()],
            ])
            ### @@ self.df['uniq_repr_nodes'] or [*umi_vignette['int_to_umi_dict'].keys()]
            # [0, 1, 2, ..., 1948]
        self.df = pd.DataFrame(
            data=res_sum,
            columns=['vignette', 'cc', 'ave_ed', 'uniq_repr_nodes'],
            index=pos_gps,
        )
        ### @@  self.df
        # vignette  ...   uniq_repr_nodes
        # 1  {'graph_adj': {0: [77, 81, 97, 153, 205, 228, ...  ...  [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,...
        # [1 rows x 3 columns]
        self.console.print('===>time for building umi graphs: {:.2f}s'.format(time.time() - umi_graph_build_stime))

        self.df['num_uniq_umis'] = self.df['uniq_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        ### @@ self.df['num_uniq_umis']
        # 1    1949
        # Name: num_uniq_umis, dtype: int64
        # self.console.print('===>start deduplication by the {} method...'.format(self.method))
        # sys.stdout.close()

    def unique(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).unique()

    def cluster(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).cluster()


    def adjacency(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).adjacency()

    def directional(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).directional()

    def mcl(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).mcl(
            **self.kwargs
        )

    def mcl_cc_all_node_umis(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).mcl_cc_all_node_umis(
            **self.kwargs
        )

    def mcl_val(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).mcl_val(
            **self.kwargs
        )

    def mcl_ed(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).mcl_ed(
            **self.kwargs
        )


if __name__ == "__main__":
    from umiche.path import to

    umiche = Gene(
        bam_fpn=to('data/RM82CLK1_S3_featurecounts_gene_sorted.bam'),
        gene_assigned_tag='XT',
        gene_is_assigned_tag='XS',
        mcl_fold_thres=1.6,
        inflat_val=1.6,
        exp_val=2,
        iter_num=100,
        ed_thres=7,
        work_dir=to('data/'),

        verbose=True,  # False
        heterogeneity=False,  # False
    )

    print(umiche.unique())
    print(umiche.cluster())
    print(umiche.adjacency())
    print(umiche.directional())
    # print(umiche.mcl())
    # print(umiche.mcl_val())
    # print(umiche.mcl_ed())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import json
import networkx as nx
import seaborn as sns
import matplotlib.pyplot as plt
from umiche.plot.gadget.Element import Element as pele
from umiche.network.Adjacency import Adjacency as netadj


class Cluster:

    def __init__(
            self,
            json_fpn=None,
    ):
        self.pele = pele()
        self.palette = self.pele.color(which='tableau', is_random=True)

        sns.set(font="Helvetica")
        sns.set_style("ticks")

        if json_fpn:
            with open(json_fpn) as fp:
                self.cnt_dict = json.load(fp)
                print(self.cnt_dict['0']['0.025'])



if __name__ == "__main__":
    from umiche.path import to

    p = Cluster(
        json_fpn=to('data/simu/mclumi/seq_errs/mcl_node_repr.json')
    )



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from umiche.util.Console import Console


class Adjacency:

    def __init__(
            self,
            verbose=True,
    ):
        self.console = Console()
        self.console.verbose = verbose

    def umi_tools(
            self,
            connected_components,
            df_umi_uniq_val_cnt,
            graph_adj,
    ):
        """
        Examples
        --------
        umi_tools adjacency wrap

        Parameters
        ----------
        connected_components
        df_umi_uniq_val_cnt
        graph_adj

        Returns
        -------

        """
        tcl = []
        cc_subs = {}
        for i, cc in connected_components.items():
            self.console.print('======>connected_components: {}'.format(cc))
            step, cc_sub = self.umi_tools_(df_umi_uniq_val_cnt=df_umi_uniq_val_cnt, cc=cc, graph_adj=graph_adj)
            tcl.append(step)
            cc_subs['cc_' + str(i)] = cc_sub
        return {
            'count': sum(tcl),
            'clusters': cc_subs,
        }

    def umi_tools_(
            self,
            df_umi_uniq_val_cnt,
            cc,
            graph_adj,
    ):
        """
        umi_tools adjacency

        Parameters
        ----------
        df_umi_uniq_val_cnt
            unique umi counts
        cc
            connected_components
        graph_adj
            the adjacency list of a graph

        Returns
        -------

        """
        cc_umi_sorted = df_umi_uniq_val_cnt.loc[df_umi_uniq_val_cnt.index.isin(cc)].sort_values(ascending=False).to_dict()
        ### @@ cc_umi_sorted
        # {'A': 456, 'E': 90, 'D': 72, 'B': 2, 'C': 2, 'F': 1}
        cc_sorted = [*cc_umi_sorted.keys()]
        ### @@ cc_sorted
        # ['A', 'E', 'D', 'B', 'C', 'F']
        visited = set()
        step = 1
        subcomponents = {}
        cc_set = set(cc_sorted)
        while cc_sorted:
            e = cc_sorted.pop(0)
            subcomponents[e] = []
            for node in graph_adj[e]:
                if node not in visited:
                    subcomponents[e].append(node)
                    self.console.print('=========>subcomponents: {}'.format(subcomponents))
            ### @@ e, subcomponents[e]
            # A ['B', 'C', 'D']
            # E []
            # D ['F']
            visited.add(e)
            ### @@ e, visited
            # A {'A'}
            # E {'B', 'A', 'E', 'C', 'D'}
            # D {'B', 'A', 'E', 'C', 'D'}
            visited.update(graph_adj[e])
            ### @@ e, visited
            # A {'B', 'D', 'A', 'C'}
            # E {'B', 'A', 'E', 'C', 'D'}
            # D {'B', 'F', 'A', 'E', 'C', 'D'}
            subcomponents[e] = graph_adj[e]
            ### @@ e, subcomponents[e]
            # A ['B', 'C', 'D']
            # E ['D']
            # D ['A', 'E', 'F']
            self.console.print('======>The ccurent ele popping out: {} {}'.format(e, visited))
            if visited == cc_set:
                # print(step)
                break
            else:
                step += 1
        ### @@ subcomponents
        # {'A': ['B', 'C', 'D'], 'E': ['D'], 'D': ['A', 'E', 'F']}
        vertex = [*subcomponents.keys()]
        ### @@ vertex
        # ['A', 'E', 'D']
        cc_sub = {}
        for k, v in subcomponents.items():
            cc_sub['node_' + str(k)] = [k]
            for i in v:
                if i not in vertex:
                    cc_sub['node_' + str(k)].append(i)
        return step, cc_sub

    def decompose(self, cc_sub_dict):
        """

        Parameters
        ----------
        cc_sub_dict

        Returns
        -------

        """
        cc_cnt = 0
        ccs = {}
        for k1, v1 in cc_sub_dict.items():
            for k2, v2 in v1.items():
                ccs[cc_cnt] = v2
                cc_cnt += 1
        return ccs


if __name__ == "__main__":
    import pandas as pd
    from umiche.deduplicate.method.Cluster import Cluster as umiclust

    p = Adjacency()

    # ### @@ data from UMI-tools
    # graph_adj = {
    #     'A': ['B', 'C', 'D'],
    #     'B': ['A', 'C'],
    #     'C': ['A', 'B'],
    #     'D': ['A', 'E', 'F'],
    #     'E': ['D'],
    #     'F': ['D'],
    # }
    # print("An adjacency list of a graph:\n{}".format(graph_adj))
    #
    # node_val_sorted = pd.Series({
    #     'A': 456,
    #     'E': 90,
    #     'D': 72,
    #     'B': 2,
    #     'C': 2,
    #     'F': 1,
    # })
    # print("Counts sorted:\n{}".format(node_val_sorted))

    ### @@ data from mine
    graph_adj = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D', 'G'],
        'F': ['D', 'G'],
        'G': ['E', 'F'],
    }
    print("An adjacency list of a graph:\n{}".format(graph_adj))

    node_val_sorted = pd.Series({
        'A': 120,
        'D': 90,
        'E': 50,
        'G': 5,
        'B': 2,
        'C': 2,
        'F': 1,
    })
    print("Counts sorted:\n{}".format(node_val_sorted))

    ccs = umiclust().cc(graph_adj=graph_adj)
    print("Connected components:\n{}".format(ccs))

    dedup_res = p.umi_tools(
        connected_components=ccs,
        df_umi_uniq_val_cnt=node_val_sorted,
        graph_adj=graph_adj
    )
    dedup_count = dedup_res['count']
    dedup_clusters = dedup_res['clusters']
    print("deduplicated count:\n{}".format(dedup_count))
    print("deduplicated clusters:\n{}".format(dedup_clusters))

    dedup_clusters_dc = p.decompose(dedup_clusters)
    print("deduplicated clusters decomposed:\n{}".format(dedup_clusters_dc))



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from umiche.network.CC import CC as gbfscc
from umiche.util.Console import Console


class Cluster:

    def __init__(
            self,
            verbose=True,
    ):
        self.console = Console()
        self.console.verbose = verbose

    def cc(
            self,
            graph_adj,
    ):
        """

        Parameters
        ----------
        graph_adj

        Returns
        -------

        """
        connected_components = list(gbfscc().deque(graph_adj))
        return {i: cc for i, cc in enumerate(connected_components)}

    def ccnx(
            self,
            edge_list,
    ):
        """

        Parameters
        ----------
        edge_list

        Returns
        -------

        """
        import networkx as nx
        G = nx.Graph()
        for edge in edge_list:
            G.add_edge(edge[0], edge[1])
        return {i: G.subgraph(cc).nodes() for i, cc in enumerate(nx.connected_components(G))}


if __name__ == "__main__":
    p = Cluster()
    # print(p.cc({0: [1,], 1: [0, 2], 2: [1]}))
    # print(p.cc({0: []}))
    graph_adj_mclumi = {
     'A': ['B', 'C', 'D'],
     'B': ['A', 'C'],
     'C': ['A', 'B'],
     'D': ['A', 'E', 'F'],
     'E': ['D', 'G'],
     'F': ['D', 'G'],
     'G': ['E', 'F'],
    }
    edge_list = [('B', 'A'), ('D', 'A'), ('C', 'B'), ('F', 'D'), ('C', 'A'), ('G', 'F'), ('E', 'D'), ('G', 'E')]
    print(p.cc(graph_adj=graph_adj_mclumi))
    print(p.ccnx(edge_list=edge_list))



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import os
from umiche.fastq.Reader import Reader as fastqread
from umiche.util.Folder import Folder as crtfolder
from umiche.util.Console import Console


class Convert:

    def __init__(
            self,
            fastq_fpn,
            bam_fpn,
            verbose=False,
    ):
        import pysam
        # import bamnostic as bs

        self.fastq_fpn = fastq_fpn
        self.bam_fpn = bam_fpn
        self.names, self.seqs, placeholders, qualities = fastqread().fromgz(fastq_fpn=fastq_fpn)
        # print(self.names)
        # print(self.seqs)

        self.console = Console()
        self.console.verbose = verbose

    def tobam(self, ):
        header = {
            'HD': {'VN': '1.0'},
            'SQ': [{'LN': 1575, 'SN': 'chr1'}, {'LN': 1584, 'SN': 'chr2'}]
        }
        crtfolder().osmkdir(DIRECTORY=os.path.dirname(self.bam_fpn))
        with pysam.AlignmentFile(self.bam_fpn, "wb", header=header) as outf:
            for name in self.names:
                a = pysam.AlignedSegment()
                a.query_name = name
                a.query_sequence = 'B'
                a.flag = 0
                a.reference_id = 0
                a.reference_start = 0
                a.mapping_quality = 0
                a.cigar = None
                a.next_reference_id = 0
                a.next_reference_start = 0
                a.template_length = 0
                a.query_qualities = None
                a.tags = (("PO", 1), )
                outf.write(a)
        return

    def tobamSimuBulk(self, ):
        header = {
            'HD': {'VN': '1.0'},
            'SQ': [{'LN': 1575, 'SN': 'chr1'}, {'LN': 1584, 'SN': 'chr2'}]
        }
        crtfolder().osmkdir(DIRECTORY=os.path.dirname(self.bam_fpn))
        with pysam.AlignmentFile(self.bam_fpn, "wb", header=header) as outf:
            for name in self.names:
                a = pysam.AlignedSegment()
                a.query_name = name
                a.query_sequence = 'B'
                a.flag = 0
                a.reference_id = 0
                a.reference_start = 0
                a.mapping_quality = 0
                a.cigar = None
                a.next_reference_id = 0
                a.next_reference_start = 0
                a.template_length = 0
                a.query_qualities = None
                a.tags = (
                    ("XT", int(name.split('-')[4])),
                    ("XS", "Assigned"),
                )
                outf.write(a)
        return

    def tobamsc(self, ):
        header = {
            'HD': {'VN': '1.0'},
            'SQ': [{'LN': 1575, 'SN': 'chr1'}, {'LN': 1584, 'SN': 'chr2'}]
        }
        crtfolder().osmkdir(DIRECTORY=os.path.dirname(self.bam_fpn))
        with pysam.AlignmentFile(self.bam_fpn, "wb", header=header) as outf:
            for name in self.names:
                asterisk_split = name.split('*')
                a = pysam.AlignedSegment()
                a.query_name = name
                a.query_sequence = 'B'
                a.flag = 0
                a.reference_id = 0
                a.reference_start = 0
                a.mapping_quality = 0
                a.cigar = None
                a.next_reference_id = 0
                a.next_reference_start = 0
                a.template_length = 0
                a.query_qualities = None
                a.tags = (
                    ("BC", int(asterisk_split[2])),
                    ("XT", int(asterisk_split[4])),
                    ("XS", "Assigned"),
                )
                outf.write(a)
        return

    def tobambulk(self, ):
        header = {
            'HD': {'VN': '1.0'},
            'SQ': [{'LN': 1575, 'SN': 'chr1'}, {'LN': 1584, 'SN': 'chr2'}]
        }
        crtfolder().osmkdir(DIRECTORY=os.path.dirname(self.bam_fpn))
        with pysam.AlignmentFile(self.bam_fpn, "wb", header=header) as outf:
            for name in self.names:
                asterisk_split = name.split('*')
                a = pysam.AlignedSegment()
                a.query_name = name
                a.query_sequence = 'B'
                a.flag = 0
                a.reference_id = 0
                a.reference_start = 0
                a.mapping_quality = 0
                a.cigar = None
                a.next_reference_id = 0
                a.next_reference_start = 0
                a.template_length = 0
                a.query_qualities = None
                a.tags = (
                    ("SP", int(asterisk_split[2])),
                    ("XT", int(asterisk_split[4])),
                    ("XS", "Assigned"),
                )
                outf.write(a)
        return


if __name__ == "__main__":
    from umiche.path import to
    p = Convert(
        # fastq_fpn=to('data/simu/transloc/trimer/bulk/seq_errs/permute_0/seq_err_0.fastq.gz'),
        # bam_fpn=to('data/simu/transloc/trimer/'),
        fastq_fpn=to('data/simu/trimer/pcr8/seq_errs/permute_0/trimmed/seq_err_18.fastq.gz'),
        bam_fpn=to('data/simu/trimer/pcr8/seq_errs/permute_0/trimmed/seq_err_18.bam'),
    )
    print(p.tobam())
    # print(p.tobamsc())
    # print(p.tobambulk())

    # print(p.tobambs())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

import pandas as pd
from umiche.fastq.Convert import Convert as fastqconverter
from umiche.trim.Template import Template as trimmer
from umiche.network.bfs.ConnectedComponent import ConnectedComponent as gbfscc
from umiche.simu.Parameter import Parameter as params

from umiche.deduplicate.MultiPos import MultiPos as deduppos
from umiche.util.Writer import Writer as fwriter
from umiche.util.Console import Console


class Standard:

    def __init__(
            self,
            scenario,
            method,
            param_fpn=None,
            is_trim=False,
            is_tobam=False,
            is_dedup=False,
            is_sv=False,
            verbose=False,
            **kwargs,
    ):
        self.scenario = scenario
        self.method = method
        self.is_sv = is_sv
        self.kwargs = kwargs
        if 'is_collapse_block' in self.kwargs.keys():
            self.is_collapse_block = self.kwargs['is_collapse_block']
        else:
            self.is_collapse_block = False
        if 'deduped_method' in self.kwargs.keys():
            self.deduped_method = self.kwargs['deduped_method']
        else:
            self.deduped_method = ''
        if 'split_method' in self.kwargs.keys():
            self.split_method = self.kwargs['split_method']
        else:
            self.split_method = ''
        if 'collapse_block_method' in self.kwargs.keys():
            self.collapse_block_method = self.kwargs['collapse_block_method']
        else:
            self.collapse_block_method = ''

        self.params = params(param_fpn=param_fpn)
        self.gbfscc = gbfscc()
        self.fwriter = fwriter()
        print('UMI homopolymer recurring pattern: {}'.format(self.params.fixed['umi_unit_pattern']))

        self.verbose = verbose
        self.console = Console()
        self.console.verbose = self.verbose

        df_dedup = pd.DataFrame()
        df_slv = pd.DataFrame()
        df_not_slv = pd.DataFrame()
        df_mono_len = pd.DataFrame()
        df_multi_len = pd.DataFrame()
        for perm_num_i in range(self.params.fixed['permutation_num']):
            print("===>Permutation number: {}".format(perm_num_i))
            dedup_arr = []
            slv_arr = []
            not_slv_arr = []
            mono_len_arr = []
            multi_len_arr = []
            for id, scenario_i in enumerate(self.params.varied[self.scenario]):
                self.console.print("======>No.{} scenario: {}".format(id+1, scenario_i))
                if self.scenario == 'pcr_nums':
                    self.fn_mark = str(scenario_i)
                elif self.scenario == 'pcr_errs':
                    self.fn_mark = str(id)
                elif self.scenario == 'seq_errs':
                    self.fn_mark = str(id)
                elif self.scenario == 'ampl_rates':
                    self.fn_mark = str(id)
                elif self.scenario == 'umi_lens':
                    self.fn_mark = str(scenario_i)
                else:
                    self.fn_mark = str(scenario_i)
                self.fn_prefix = self.params.file_names[self.scenario] + self.fn_mark
                self.fastq_location = self.params.work_dir + self.scenario + '/permute_' + str(perm_num_i) + '/'
                if is_trim:
                    self.console.print("======>fastq is being trimmed.")
                    self.params.trimmed['fastq']['fpn'] = self.fastq_location + self.fn_prefix + '.fastq.gz'
                    self.params.trimmed['fastq']['trimmed_fpn'] = self.fastq_location + 'trimmed/' + self.fn_prefix + '.fastq.gz'
                    umitrim_parser = trimmer(params=self.params.trimmed, verbose=self.verbose)
                    df = umitrim_parser.todf()
                    umitrim_parser.togz(df)
                if is_tobam:
                    self.console.print("======>fastq converter to bam is being used.")
                    fastqconverter(
                        fastq_fpn=self.fastq_location + 'trimmed/' + self.fn_prefix + '.fastq.gz',
                        bam_fpn=self.fastq_location + 'trimmed/bam/' + self.fn_prefix + '.bam',
                    ).tobam()
                if is_dedup:
                    self.console.print("======>reads are being deduplicated.")
                    if self.is_collapse_block:
                        if self.deduped_method == '' and self.split_method == '':
                            plus = ''
                        else:
                            plus = self.deduped_method + '_' + self.split_method
                        bam_fpn = self.fastq_location + 'trimmed/bam/' + plus + '/' + self.fn_prefix + '.bam'
                    else:
                        bam_fpn = self.fastq_location + 'trimmed/bam/' + self.fn_prefix + '.bam'
                    if self.method == 'set_cover' or self.method == 'majority_vote':
                        self.is_build_graph = False
                    else:
                        self.is_build_graph = True
                    # print(self.is_collapse_block)
                    dedup_ob = deduppos(
                        bam_fpn=bam_fpn,
                        pos_tag='PO',
                        mcl_fold_thres=self.params.dedup['mcl_fold_thres'],
                        inflat_val=self.params.dedup['inflat_val'],
                        exp_val=self.params.dedup['exp_val'],
                        iter_num=self.params.dedup['iter_num'],
                        ed_thres=self.params.dedup['ed_thres'],
                        work_dir=self.params.work_dir,
                        sv_interm_bam_fpn=self.fastq_location + 'trimmed/bam/' + self.method + '_' + self.split_method + '/' + self.fn_prefix + '.bam',
                        heterogeneity=False, # False True
                        is_build_graph=self.is_build_graph,
                        is_collapse_block=self.is_collapse_block,
                        umi_unit_pattern=self.params.fixed['umi_unit_pattern'],
                        split_method=self.split_method,
                        collapse_block_method=self.collapse_block_method,
                        verbose=self.verbose,
                        # **self.kwargs
                    )
                    df = self.tool(dedup_ob)[self.method]()
                    if self.method == "unique":
                        print("============>No.{}, dedup cnt: {}".format(id, df.num_uniq_umis.values[0]))
                        dedup_arr.append(df.num_uniq_umis.values[0])
                    elif self.method == "set_cover":
                        print("============>No.{}, dedup cnt: {}".format(id, df.dedup_cnt.values[0]))
                        # print("============>No.{}, num solved: {}".format(id, df.num_solved.values[0]))
                        # print("============>No.{}, num not solved: {}".format(id, df.num_not_solved.values[0]))
                        dedup_arr.append(df.dedup_cnt.values[0])
                        slv_arr.append(df.num_solved.values[0])
                        not_slv_arr.append(df.num_not_solved.values[0])
                        mono_len_arr.append(df.monomer_umi_len.values[0])
                        multi_len_arr.append(df.multimer_umi_len.values[0])
                    else:
                        print("============>No.{}, dedup cnt: {}".format(id, df.dedup_cnt.values[0]))
                        dedup_arr.append(df.dedup_cnt.values[0])
            df_dedup['pn' + str(perm_num_i)] = dedup_arr
            if self.method == 'set_cover':
                df_slv['pn' + str(perm_num_i)] = slv_arr
                df_not_slv['pn' + str(perm_num_i)] = not_slv_arr
                df_mono_len['pn' + str(perm_num_i)] = mono_len_arr
                df_multi_len['pn' + str(perm_num_i)] = multi_len_arr
            # print(df_dedup)
        if not self.is_collapse_block:
            remark = 'collapse_block'
        else:
            remark = 'no_collapse_block'
        self.sv_dedup_dir = self.params.work_dir + self.scenario + '/' + remark + '/'
        from umiche.util.Folder import Folder as crtfolder
        crtfolder().osmkdir(DIRECTORY=self.sv_dedup_dir)
        # self.fwriter.generic(
        #     df=df_dedup,
        #     sv_fpn=self.sv_dedup_dir + self.method + '_dedupby_' + self.deduped_method + '_splitby_' + self.split_method + '_collblockby_' + self.collapse_block_method + '_dedup.txt',
        #     header=True,
        # )
        # if self.method == 'set_cover':
            # self.fwriter.generic(
            #     df=df_slv,
            #     sv_fpn=self.sv_dedup_dir + self.method + '_solved_' + self.split_method + '_dedup.txt',
            #     header=True,
            # )
            # self.fwriter.generic(
            #     df=df_not_slv,
            #     sv_fpn=self.sv_dedup_dir + self.method + '_not_solved_' + self.split_method + '_dedup.txt',
            #     header=True,
            # )
            # self.fwriter.generic(
            #     df=df_mono_len,
            #     sv_fpn=self.sv_dedup_dir + self.method + '_mono_len_' + self.split_method + '_dedup.txt',
            #     header=True,
            # )
            # self.fwriter.generic(
            #     df=df_multi_len,
            #     sv_fpn=self.sv_dedup_dir + self.method + '_multi_len_' + self.split_method + '_dedup.txt',
            #     header=True,
            # )

    def tool(self, dedup_ob):
        return {
            'unique': dedup_ob.unique,
            'cluster': dedup_ob.cluster,
            'adjacency': dedup_ob.adjacency,
            'directional': dedup_ob.directional,
            'mcl': dedup_ob.mcl,
            'mcl_val': dedup_ob.mcl_val,
            'mcl_ed': dedup_ob.mcl_ed,
            'mcl_cc_all_node_umis': dedup_ob.mcl_cc_all_node_umis,
            # 'dbscan_seq_onehot': dedup_ob.dbscan_seq_onehot,
            # 'birch_seq_onehot': dedup_ob.birch_seq_onehot,
            # 'hdbscan_seq_onehot': dedup_ob.hdbscan_seq_onehot,
            # 'aprop_seq_onehot': dedup_ob.aprop_seq_onehot,
            'set_cover': dedup_ob.set_cover,
            'majority_vote': dedup_ob.majority_vote,
        }


if __name__ == "__main__":
    from umiche.path import to

    p = Standard(
        # scenario='pcr_nums',
        # scenario='pcr_errs',
        scenario='seq_errs',
        # scenario='ampl_rates',
        # scenario='umi_lens',

        # method='unique',
        # method='cluster',
        # method='adjacency',
        # method='directional',
        # method='mcl',
        # method='mcl_val',
        # method='mcl_ed',
        method='set_cover',
        # method='majority_vote',

        # is_trim=True,
        # is_tobam=False,
        # is_dedup=False,

        # is_trim=False,
        # is_tobam=True,
        # is_dedup=False,

        is_trim=False,
        is_tobam=False,
        is_dedup=True,

        # @@ for directional on multimer umis deduplicated by set_cover
        is_collapse_block=False,
        deduped_method='set_cover',
        split_method='split_to_all', # split_to_all split_by_mv

        # @@ for directional on multimer umis deduplicated by majority_vote
        # is_collapse_block=False,
        # deduped_method='majority_vote',
        # split_method='',

        # @@ for directional but on monomer umis of set_cover or majority_vote
        # is_collapse_block=True, # True False
        # collapse_block_method='take_by_order', # majority_vote take_by_order
        # deduped_method='set_cover', # majority_vote set_cover
        # split_method='split_by_mv', # split_to_all split_by_mv

        # @@ for directional but on monomer umis without other methods.
        # is_collapse_block=True,  # True False
        # collapse_block_method='majority_vote',  # majority_vote take_by_order
        # deduped_method='',  # majority_vote set_cover
        # split_method='',  # split_to_all split_by_mv

        # param_fpn=to('data/params_dimer.yml'),
        param_fpn=to('data/params_trimer.yml'),
        # param_fpn=to('data/params.yml'),

        verbose=False, # True False
    )


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"


import numpy as np
import matplotlib.pyplot as plt
from umiche.plot.boxplot.Style import Style


class boxplot:

    def __init__(self, ):
        pass

    def draw1(
            self,
            data_dict,
    ):
        """
        ..  Description:
            ------------
            data_dict, which is a python dictionary, is subject to the data structure below:
            {
                {method_1: {val_dict_arr_1d_1: val_dict_containing_arr_1d_1}},
                {method_2: {val_dict_arr_1d_2: val_dict_containing_arr_1d_2}},
                {method_3: {val_dict_arr_1d_3: val_dict_containing_arr_1d_3}},
                ...
                {method_m: {val_dict_arr_1d_m: val_dict_containing_arr_1d_m}},
            }

            {
                'real': {
                    'val_dict_arr_1d':
                    {'CD8+ Cytotoxic T': array([31, 33, 17, ..., 30, 27, 23], dtype=int64),
                     'CD8+/CD45RA+ Naive Cytotoxic': array([31, 33, 17, ..., 27, 23, 21], dtype=int64),
                     'CD4+/CD45RO+ Memory': array([31, 33, 17, ..., 30, 23, 27], dtype=int64),
                     'CD19+ B': array([31, 33, 17, ..., 35, 27, 21], dtype=int64),
                     'CD4+/CD25 T Reg': array([31, 17, 33, ..., 30, 23, 27], dtype=int64),
                     'CD56+ NK': array([31, 17,  8, ..., 34,  8, 20], dtype=int64),
                     'CD4+ T Helper2': array([31, 33, 17, ..., 30, 27, 23], dtype=int64),
                     'CD4+/CD45RA+/CD25- Naive T': array([31, 33, 17, ..., 27, 21, 23], dtype=int64),
                     'CD34+': array([ 8,  0, 17, ..., 36,  0, 26], dtype=int64),
                     'Dendritic': array([31,  8,  3, ..., 26,  9, 35], dtype=int64),
                     'CD14+ Monocyte': array([ 3,  8, 31, ..., 35, 26, 12], dtype=int64)}
                }
                'method_2': {
                    ...
                }
            }

        ..  Example:
            --------

        :param data_dict: 2d dict
        :return: boxplot
        """
        key_1st = next(iter(data_dict))
        num_plot_pos = len(data_dict[key_1st]['val_dict_arr_1d'])
        num_plot_per_pos = len(data_dict)
        meanpointprops = {
            'marker': 'o',
            'markeredgecolor': 'grey',
            'markerfacecolor': 'grey',
            'markersize': 6
        }
        method_val_dict = {}
        method_pos_dict = {}
        offset_between_bplot_per_pos = 0.6
        for i_met, (met_key, v) in enumerate(data_dict.items()):
            method_pos_dict[met_key] = np.arange(num_plot_pos) * num_plot_per_pos - i_met * offset_between_bplot_per_pos
            method_val_dict[met_key] = [arr_1d for arr_1d in v['val_dict_arr_1d'].values()]
        print(method_val_dict)
        print(method_pos_dict)

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(14, 7))
        bplot_handles = []
        # bplot_handles store methods with size equal to the number of the methods
        for val_key, pos_key in zip(method_val_dict, method_pos_dict):
             val_arr_2d = method_val_dict[val_key]
             pos_arr_1d = method_pos_dict[pos_key]
             bplot_handles.append(ax.boxplot(
                x=val_arr_2d,
                positions=pos_arr_1d,
                showmeans=True,
                meanprops=meanpointprops,
                patch_artist=True,
                showfliers=False, # hide outliers, the same as sym=''
            ))
        x_ticks = np.arange(0, num_plot_pos * num_plot_per_pos, num_plot_per_pos)
        x_tick_labels = [*data_dict[key_1st]['val_dict_arr_1d'].keys()]
        ax.set_xticks(x_ticks)
        ax.set_xticklabels(x_tick_labels, rotation=30, ha='right', fontsize=9)

        from matplotlib import rcParams
        rcParams['font.family'] = 'sans-serif'
        rcParams['font.sans-serif'] = ['Tahoma']
        ax.set_xlabel('Cell cluster', fontsize=18)
        ax.set_ylabel('Number of UMI counts', fontsize=18)
        ax.yaxis.grid(True)
        ax.spines['right'].set_color('none')
        ax.spines['top'].set_color('none')

        fig.subplots_adjust(
            bottom=0.26,
            # top=0.92, left=0.10, right=0.95, hspace=0.25, wspace=0.17
        )
        palette = [
            'magenta',
            'skyblue',
            'gray',
            'sienna',
            'crimson',
        ]
        Style().plain(boxplot_handles=bplot_handles, palette=palette)

        # add legend to the boxplot
        ax.legend(
            handles=[i["boxes"][0] for i in bplot_handles],
            labels=[*data_dict.keys()],
            bbox_to_anchor=(0.5, 1.15),
            loc=9,
            ncol=7,
        )
        # fig.tight_layout()
        plt.show()
        return 0

    def draw2(self, ):
        return




__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from umiche.simu.Parameter import Parameter as params
from umiche.deduplicate.io.Stat import Stat as dedupstat
from umiche.plot.gadget.Transmitter import Transmitter as transmitter

from umiche.util.Reader import Reader as freader


class DedupMultiple:

    def __init__(
            self,
            scenarios: Dict,
            methods: Dict,
            umi_gt_cnt: int = 50,
            param_fpn : str = None,
    ):
        self.scenarios = scenarios
        self.methods = methods
        self.umi_gt_cnt = umi_gt_cnt
        self.param_fpn = param_fpn
        self.freader = freader()
        self.params = params(param_fpn=self.param_fpn)

        self.dedupstat = dedupstat(
            scenarios=self.scenarios,
            methods=self.methods,
            param_fpn=self.param_fpn,
            is_trans=False,
        )

        self.df_dedup = self.dedupstat.df_runtime

        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def line4(
            self,
            num_img_row=2,
            num_img_col=2,
    ):
        print(self.df_dedup)
        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(13.1, 9), sharey=False, sharex=False)
        palette = sns.color_palette("Paired")
        for n, (scenario, scenario_formal) in enumerate(self.scenarios.items()):
            self.df_sce = self.df_dedup[self.df_dedup['scenario'] == scenario_formal]
            for j, (method, method_formal) in enumerate(self.methods.items()):
                self.df_sce_met = self.df_sce[self.df_sce['method'] == method_formal]

                if int(n / num_img_col) == num_img_row - 1:
                    xl_mark = True
                elif (int(n / num_img_col) == num_img_row - 2) and (
                        n % num_img_col >= (num_img_col - num_img_col * num_img_row + len([1]))):
                    xl_mark = True
                else:
                    xl_mark = False

                print(int(n / num_img_col), n % num_img_col)

                ax[int(n / num_img_col), n % num_img_col].errorbar(
                    x=self.df_sce_met.index,
                    y=self.df_sce_met['mean'].values,
                    yerr=[self.df_sce_met['mean-min'], self.df_sce_met['max-mean']],
                    fmt='o',
                    # alpha=0.6,
                    # ecolor=palette[j],
                    color=palette[j],
                    linestyle='-',
                    linewidth=2,
                    elinewidth=1,
                    capsize=3,
                    markersize=6,
                    markeredgewidth=1.2,
                    markeredgecolor=palette[j],
                    markerfacecolor='none',
                    # label=method_formal,
                )

                self.line_gadget(
                    ax=ax[int(n / num_img_col), n % num_img_col],
                    x=self.df_sce_met.index,
                    y=self.df_sce_met['mean'].values,
                    line_color=palette[j],
                    label=method_formal,  # " ".join(ds_key.split("_"))
                    linewidth=2,
                    marker="o",
                    marker_size=6,
                    marker_edge_width=1.2,
                    marker_face_color='none',
                    decoration_mark=True,
                    xl_mark=xl_mark,
                    yl_mark=True if n % num_img_col == 0 else False,
                    title='{}'.format(scenario_formal),  # '{}'.format(scenario_formal)
                    x_label=None,  # 'Position'
                    # title='{}'.format(scenario_formal), # '{}'.format(scenario_formal)
                    # x_label='PCR cycles', # 'Position'
                    # title=None,  # '{}'.format(scenario_formal)
                    # x_label='PCR cycles',  # 'Position'
                    y_label='Runtime (s)',
                    title_fs=16,
                    x_label_rotation=20,
                    x_label_rotation_align='right',
                    x_ticks=(np.arange(self.df_sce_met.index.shape[0])).tolist(),
                    x_ticklabels=self.df_sce_met['metric'].values.tolist(),
                    x_ticklabel_fs=11,
                    x_label_fs=14,
                    y_label_fs=18,
                    legend_fs=11,
                    legend_loc=None,
                )
        for i in range(num_img_col * num_img_row - (n + 1)):
            ax[num_img_row - 1, num_img_col - 1 - i].set_visible(False)
        plt.subplots_adjust(
            top=0.96,
            bottom=0.06,
            left=0.06,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        plt.show()
        return

    def line(
            self,
            num_img_row=1,
            num_img_col=1,
    ):
        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(6.5, 5.5), sharey=False, sharex=False)
        palette = sns.color_palette("Paired")[:]
        # palette = sns.color_palette("Set3")[3:]
        for n, (scenario, scenario_formal) in enumerate(self.scenarios.items()):
            self.df_sce = self.df_dedup[self.df_dedup['scenario'] == scenario_formal]
            for j, (method, method_formal) in enumerate(self.methods.items()):
                self.df_sce_met = self.df_sce[self.df_sce['method'] == method_formal]

                if int(n / num_img_col) == num_img_row - 1:
                    xl_mark = True
                elif (int(n / num_img_col) == num_img_row - 2) and (
                        n % num_img_col >= (num_img_col - num_img_col * num_img_row + len([1]))):
                    xl_mark = True
                else:
                    xl_mark = False

                # ax[int(n / num_img_col), n % num_img_col].errorbar(
                ax.errorbar(
                    x=self.df_sce_met.index,
                    y=self.df_sce_met['mean'].values,
                    yerr=[self.df_sce_met['mean-min'], self.df_sce_met['max-mean']],
                    fmt='o',
                    # alpha=0.6,
                    # ecolor=palette[j],
                    color=palette[j],
                    linestyle='-',
                    linewidth=2,
                    elinewidth=1,
                    capsize=3,
                    markersize=6,
                    markeredgewidth=1.2,
                    markeredgecolor=palette[j],
                    markerfacecolor='none',
                    label=method_formal,
                )
                print('asdas',self.df_sce_met.index)
                self.line_gadget(
                    ax=ax,
                    x=self.df_sce_met.index,
                    y=self.df_sce_met['mean'].values,
                    line_color=palette[j],
                    # label=method_formal,  # " ".join(ds_key.split("_"))
                    linewidth=2,
                    marker="o",
                    marker_size=6,
                    marker_edge_width=1.2,
                    marker_face_color='none',
                    decoration_mark=True,
                    xl_mark=xl_mark,
                    yl_mark=True if n % num_img_col == 0 else False,
                    # title='{}'.format(scenario_formal),
                    title_fs=16,
                    x_label='{}'.format(scenario_formal),  # 'Position'
                    y_label='Runtime (s)',
                    x_label_rotation=30,
                    x_label_rotation_align='right',
                    x_ticks=(np.arange(self.df_sce_met.index.shape[0])).tolist(),
                    x_ticklabels=self.df_sce_met['metric'].values.tolist(),
                    x_ticklabel_fs=12,
                    x_label_fs=18,
                    y_label_fs=18,
                    legend_fs=14,
                    legend_loc=None,
                )
                ax.spines['bottom'].set_linewidth(2)
                ax.spines['left'].set_linewidth(2)

        for i in range(num_img_col * num_img_row - (n + 1)):
            ax[num_img_row - 1, num_img_col - 1 - i].set_visible(False)
        plt.subplots_adjust(
            top=0.96,
            bottom=0.16,
            left=0.12,  # 0.14 0.12
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        # plt.savefig('./umitools mets.eps', format='eps')
        # plt.savefig('./mv+sc.eps', format='eps')
        plt.show()
        return

    @transmitter(type="line", task=None)
    def line_gadget(*args, **kwargs):
        return kwargs

    @transmitter(type="line_scatter", task=None)
    def line_scatter_gadget(*args, **kwargs):
        return kwargs


if __name__ == "__main__":
    from umiche.path import to

    p = DedupMultiple(
        scenarios={
            # 'pcr_nums_gene': '10 samples, 10 genes (34496 counts)',
            # 'pcr_nums_sc': '10 cells, 10 genes (1828 counts)',

            'pcr_nums': 'PCR cycle',
            'pcr_errs': 'PCR error',
            'seq_errs': 'Sequencing error',
            'seq_deps': 'Sequencing depth',

            # 'ampl_rates': 'Amplification rate',

            # 'pcr_nums_17-20': 'high PCR cycle',
            # 'pcr_nums_21-24': 'high PCR cycle1',
        },

        methods={
            'tree': 'bfTree',
            'minnow': 'spTree',
            'mutation_table_minimum': 'ErrorTable-s',
            'mutation_table_complete': 'ErrorTable-m',
            'err1d': 'ReadWhole',
            'err2d': 'ReadPer',
        },

        param_fpn=to('data/tresor/params.yml'),
    )

    # p.line()
    p.line4()